unit xmlrpc_cli;
{ xmlrpc_cli unit
  XML-RPC Client class implementation.
  PRE-PRE-BETA QUALITY!!
}

interface
uses
    JclUnicode,LibXmlParser,utf8xmlpars,
  Windows, Messages, SysUtils, Classes, Controls,
  Dialogs, StdCtrls,IdBaseComponent, IdComponent, IdTCPConnection,IdTCPClient, IdHTTP;

type
  //see XML-RPC spec
  TXMLRPCValueType = (rvI4,rcBoolean,rvString,rvDouble,rvDateTime_iso8601,rvBase64,rvArray,rvStruct);

  //base class
  TXMLRPCValue = class(TObject)
  public
  end;

  TXMLRPCRequest = class(TObject)
  private
   FRawRequestXML: string;
   FCodeConversionEnabled,FNormalize: Boolean;
   function GetMethodName: string;
  public
   property RequestMethodName: string read GetMethodName;
   property Normalize: Boolean read FNormalize write FNormalize;
   function GetAsXML: string; overload;
   function GetAsXML(ACodeTranslation: Boolean): string; overload;
   constructor Create(ARequestXml: string; ACodeConversionEnabled,ANormalize: Boolean);
   destructor Destroy; override;
  end;

  //XML-RPC Response base class
  TXMLRPCResponse = class(TObject)
  private
   FRawResponseXML: string;
   FCodeConversionEnabled,FNormalize: Boolean;
   function GetIsFault: Boolean;
   function GetFaultCode: integer;
   function GetFaultString: string;
   //parse faultCode and return faultCode,faultString
   procedure ParseFaultResponse(AFaultResponseXML: string; var AFaultCode: integer;var AFaultString : string);
  public
   property Normalize: Boolean read FNormalize write FNormalize;
   property IsFault: Boolean read GetIsFault;
   property FaultCode: integer read GetFaultCode;
   property FaultString: string read GetFaultString;
   function GetAsXML: string; overload;
   function GetAsXML(ACodeTranslation: Boolean): string; overload;
   constructor Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
   destructor Destroy; override;
  end;

  //Only for system.listMethods
  TXMLRPCListMethods = class(TXMLRPCResponse)
  private
   FValues: TStrings;
   procedure ParseListMethods;
  public
   property Values: TStrings read FValues;
   constructor Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
   destructor Destroy; override;
  end;

  //Only for system.methodHelp
  TXMLRPCMethodHelp = class(TXMLRPCResponse)
  private
   FHelp: string;
   procedure ParseMethodHelp;
  public
   property Help: string read FHelp;
   constructor Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
   destructor Destroy; override;
  end;

  //Only for system.methodSignature
  TXMLRPCMethodSignature = class(TXMLRPCResponse)
  private
   FParams: TStrings;
   procedure ParseMethodSignature;
  public
   property Params: TStrings read FParams;
   constructor Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
   destructor Destroy; override;
  end;

  //AStrings is valid XML-RPC Request?
  function IsValidXMLRPCRequest(AStrings: TStrings): Boolean;

implementation

{------------------------------------------------------------------------------}
{ TXMLRPCRequest                                                              }
{------------------------------------------------------------------------------}
constructor TXMLRPCRequest.Create(ARequestXML: string; ACodeConversionEnabled,ANormalize: Boolean);
begin
     FRawRequestXML := ARequestXML;
     FCodeConversionEnabled := ACodeConversionEnabled;
     FNormalize := ANormalize;
end;

destructor TXMLRPCRequest.Destroy;
begin
   inherited Destroy;
end;

function TXMLRPCRequest.GetMethodName: string;
begin
    Result := '';
    with TUtf8ToWideXMLParser.Create do
    begin
     CodeConversionEnabled := FCodeConversionEnabled;
     LoadFromBuffer(PChar(FRawRequestXml));
     StartScan;
     while Scan do
      if (CurPartType=ptContent) then
      begin
       if (CurName='methodName') then
       begin
        Result := CurContent;
        Break;
       end;
      end;
    end;
end;

function TXMLRPCRequest.GetAsXML: string;
begin
     if (FCodeConversionEnabled) then
      Result := ConvertWs(Utf8ToWideString(FRawRequestXML),Normalize)
     else Result := ConvertWs(FRawRequestXML,Normalize);
end;

function TXMLRPCRequest.GetAsXML(ACodeTranslation: Boolean): string;
begin
    if (ACodeTranslation) then Result := GetAsXML
    else Result := ConvertWs(FRawRequestXML,Normalize);
end;


{------------------------------------------------------------------------------}
{ TXMLRPCResponse                                                              }
{------------------------------------------------------------------------------}
constructor TXMLRPCResponse.Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
begin
     FRawResponseXML := AResponseXml;
     FCodeConversionEnabled := ACodeConversionEnabled;
     FNormalize := ANormalize;
end;

destructor TXMLRPCResponse.Destroy;
begin
   inherited Destroy;
end;

function TXMLRPCResponse.GetAsXML: string;
begin
     if (FCodeConversionEnabled) then
      Result := ConvertWs(Utf8ToWideString(FRawResponseXML),Normalize)
     else Result := ConvertWs(FRawResponseXML,Normalize);
end;

function TXMLRPCResponse.GetAsXML(ACodeTranslation: Boolean): string;
begin
    if (ACodeTranslation) then Result := GetAsXML
    else Result := ConvertWs(FRawResponseXML,Normalize);
end;

//if faultCode exists return true
function TXMLRPCResponse.GetIsFault: Boolean;
begin
    Result := false;
    with TUtf8ToWideXMLParser.Create do
    begin
     CodeConversionEnabled := FCodeConversionEnabled;
     LoadFromBuffer(PChar(FRawResponseXml));
     StartScan;
     while Scan do
      if (CurPartType=ptContent) then
      begin
       if (CurContent='faultCode') then
       begin
        Result := true;
        Break;
       end;
      end;
    end;
end;

//parse fault response, return faultCode and faultString
procedure TXMLRPCResponse.ParseFaultResponse(AFaultResponseXML: string; var AFaultCode: integer;var AFaultString : string);
var
   Parser: TUtf8ToWideXMLParser;
begin
     Parser := TUtf8ToWideXMLParser.Create;
     try
      with Parser do
      begin
       CodeConversionEnabled := FCodeConversionEnabled;

       LoadFromBuffer(PChar(AFaultResponseXml));
       StartScan;
       while Scan do
        if (CurPartType=ptContent) then
        begin
         if (CurName='int') then AFaultCode := StrToIntDef(CurContent,-1);
         if (CurName='string') then AFaultString := CurContent;
        end;
      end;
     finally
      Parser.Free;
     end;
end;

function TXMLRPCResponse.GetFaultCode: integer;
var
   AFaultCode: integer;
   AFaultString: string;
begin
    Result := 0;
    if not self.IsFault then exit;
    ParseFaultResponse(FRawResponseXml,AFaultCode,AFaultString);
    Result := AFaultCode;
end;

function TXMLRPCResponse.GetFaultString: string;
var
   AFaultCode: integer;
   AFaultString: string;
begin
    Result := '';
    if not self.IsFault then exit;
    ParseFaultResponse(FRawResponseXml,AFaultCode,AFaultString);
    Result := AFaultString;
end;

{------------------------------------------------------------------------------}
{ TXMLRPCListMethods system.listMethodsp                                    }
{------------------------------------------------------------------------------}
constructor TXMLRPCListMethods.Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
begin
   inherited Create(AResponseXML,ACodeConversionEnabled,ANormalize);
     FValues := TStringList.Create;
     ParseListMethods;
end;

destructor TXMLRPCListMethods.Destroy;
begin
     FValues.Free;
   inherited Destroy;
end;

procedure TXMLRPCListMethods.ParseListMethods;
var
   Parser: TUtf8ToWideXMLParser;
begin
     Parser := TUtf8ToWideXMLParser.Create;
     try
      with Parser do
      begin
       CodeConversionEnabled := FCodeConversionEnabled;
       LoadFromBuffer(PChar(FRawResponseXml));
       StartScan;
       while Scan do
        if (CurPartType=ptContent) then
        begin
         if (CurName='string') then
         begin
          FValues.Add(CurContent);
         end;
        end;
      end;
     finally
      Parser.Free;
     end;
end;


{------------------------------------------------------------------------------}
{ TXMLRPCMethodHelp system.methodHelpp                                      }
{------------------------------------------------------------------------------}
constructor TXMLRPCMethodHelp.Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
begin
   inherited Create(AResponseXML,ACodeConversionEnabled,ANormalize);
     FHelp := '';
     ParseMethodHelp;
end;

destructor TXMLRPCMethodHelp.Destroy;
begin
   inherited Destroy;
end;

procedure TXMLRPCMethodHelp.ParseMethodHelp;
var
   Parser: TUtf8ToWideXMLParser;
begin
     Parser := TUtf8ToWideXMLParser.Create;
     try
      with Parser do
      begin
       CodeConversionEnabled := FCodeConversionEnabled;
       LoadFromBuffer(PChar(FRawResponseXml));
       StartScan;
       while Scan do
        if (CurPartType=ptContent) then
        begin
         if (CurName='string') then
         begin
          FHelp := CurContent;
         end;
        end;
      end;
     finally
      Parser.Free;
     end;
end;


{------------------------------------------------------------------------------}
{ TXMLRPCMethodSignature system.methodSignaturep                            }
{------------------------------------------------------------------------------}
constructor TXMLRPCMethodSignature.Create(AResponseXml: string; ACodeConversionEnabled,ANormalize: Boolean);
begin
   inherited Create(AResponseXML,ACodeConversionEnabled,ANormalize);
     FParams := TStringList.Create;
     ParseMethodSignature;
end;

destructor TXMLRPCMethodSignature.Destroy;
begin
     FParams.Free;
   inherited Destroy;
end;

procedure TXMLRPCMethodSignature.ParseMethodSignature;
var
   Parser: TUtf8ToWideXMLParser;
begin
     Parser := TUtf8ToWideXMLParser.Create;
     try
      with Parser do
      begin
       CodeConversionEnabled := FCodeConversionEnabled;
       LoadFromBuffer(PChar(FRawResponseXml));
       StartScan;
       while Scan do
        if (CurPartType=ptContent) then
        begin
         if (CurName='string') then
         begin
          FParams.Add(CurContent);
         end;
        end;
      end;
     finally
      Parser.Free;
     end;
end;


//Is valid XML-RPC Request?
function IsValidXMLRPCRequest(AStrings: TStrings): Boolean;
var
   AMethodName: string;
   AHasMethodCallStart,AHasMethodNameStart: Boolean;
begin
    AHasMethodCallStart := false;
    AHasMethodNameStart := false;
    AMethodName := '';
    with TUtf8ToWideXMLParser.Create do
    begin
     CodeConversionEnabled := false;
     LoadFromBuffer(PChar(AStrings.Text));
     StartScan;
     while Scan do
     begin
      if (CurPartType=ptContent) and (CurName='methodName') then AMethodName := CurContent;
      if (CurName='methodCall') then AHasMethodCallStart := true;
      if (CurName='methodName') then AHasMethodNameStart := true;
     end;
    end;
    Result := (AMethodName<>'') and AHasMethodCallStart and AHasMethodNameStart;
end;

end.
